unit ImgFilter;

interface

uses Windows;

type
  TImgInfo = record
    Width: Integer;
    Height: Integer;
    Bpp: Integer;
    MimeType: string;
    FormatName: string;
    Comment: string;
  end;

  TImgFilter = class(TObject)
    Handle: THandle;
    FName: string;
    FErrorMsg: string;
    function Test(Buf: PChar; BufSize: Longint): Boolean; virtual; abstract;
    function GetInfo(Buf: PChar; BufSize: Longint; var ImgInfo: TImgInfo): Boolean; virtual; abstract;
    function Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar;
      OutBufSize: Longint; BitmapInfo: PBitmapInfo): Boolean; virtual; abstract;
    property ErrorMsg: string read FErrorMsg;
    property Name: string read FName;
  end;

  TDibFilter = class(TImgFilter)
    constructor Create;
    function Test(Buf: PChar; BufSize: Longint): Boolean; override;
    function GetInfo(Buf: PChar; BufSize: Longint; var ImgInfo: TImgInfo): Boolean; override;
    function Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar;
      OutBufSize: Longint; BitmapInfo: PBitmapInfo): Boolean; override;
  end;

implementation

const
  MaxBmiSize = SizeOf(TBitmapInfoHeader)+SizeOf(TRGBQuad)*(256+3);

function IsOS2Dib(Buf: PChar): Boolean;
begin
  with PBitmapInfo(@Buf[SizeOf(TBitmapFileHeader)])^ do
    Result := bmiHeader.biSize = SizeOf(TBitmapCoreHeader);
end;

procedure CoreInfoToBitmapInfo(BmInfo: PBitmapInfo; BmCore: PBitmapCoreInfo);
var
  i: Integer;
begin
  with BmInfo^ do
  begin
    FillChar(bmiHeader, SizeOf(bmiHeader), 0);
    with bmiHeader do
    begin
      biSize := SizeOf(bmiHeader);
      biWidth := BmCore.bmciHeader.bcWidth;
      biHeight := BmCore.bmciHeader.bcHeight;
      biPlanes := BmCore.bmciHeader.bcPlanes;
      biBitCount := BmCore.bmciHeader.bcBitCount;
      biCompression := BI_RGB;
    end;
    if (bmiHeader.biBitCount = 1) or (bmiHeader.biBitCount = 4)
       or (bmiHeader.biBitCount = 8) then
    begin
      for i := 0 to 1 shl bmiHeader.biBitCount - 1 do
      begin
        bmiColors[i].rgbBlue := BmCore.bmciColors[i].rgbtBlue;
        bmiColors[i].rgbGreen := BmCore.bmciColors[i].rgbtGreen;
        bmiColors[i].rgbRed := BmCore.bmciColors[i].rgbtRed;
      end;
    end;
  end;
end;

function IsRLE(Buf: PChar): Boolean;
begin
  with PBitmapInfo(@Buf[SizeOf(TBitmapFileHeader)])^ do
    Result := (bmiHeader.biCompression = BI_RLE8) or (bmiHeader.biCompression = BI_RLE4);
end;

procedure GetBitmapInfo(Buf: PChar; BitmapInfo: PBitmapInfo);
begin
  if IsOS2Dib(Buf) then
    CoreInfoToBitmapInfo(BitmapInfo, PBitmapCoreInfo(@Buf[SizeOf(TBitmapFileHeader)]))
  else begin
    Move(Buf[SizeOf(TBitmapFileHeader)], BitmapInfo^, MaxBmiSize);
    if IsRLE(Buf) then
      with BitmapInfo.bmiHeader do
      begin
        biBitCount := 24;
        biCompression := BI_RGB;
        biSizeImage := 0;
      end;
  end;
end;

function DecodeRLE(InBuf: PChar; InBufSize: Integer; OutBuf: PChar; OutBufSize: Integer): Boolean;
var
  dc: HDC;
  wnd: HWND;
  bmp: HBITMAP;
  bmi: PBitmapInfo;
  offbits, pitch: Integer;
  decbmi: TBitmapInfo;
begin
  offbits := PBitmapFileHeader(InBuf).bfOffBits;
  bmi := PBitmapInfo(@InBuf[SizeOf(TBitmapFileHeader)]);
  // biSizeImage sȏꍇC (TODO:CƂʒm)
  if (bmi.bmiHeader.biSizeImage = 0) and (bmi.bmiHeader.biCompression <> BI_RGB) then
    bmi.bmiHeader.biSizeImage := InBufSize - offbits;

  wnd := GetDesktopWindow;
  dc := GetDC(wnd);
  bmp := CreateDIBitmap(dc, bmi.bmiHeader, CBM_INIT, PChar(@InBuf[offbits]), bmi^, DIB_RGB_COLORS);
  try
    decbmi := bmi^;
    with decbmi.bmiHeader do
    begin
      biBitCount := 24;
      biCompression := BI_RGB;
      biSizeImage := 0;
    end;
    pitch := (decbmi.bmiHeader.biWidth * 24 shr 3 + 3) and not 3;
    if (pitch * decbmi.bmiHeader.biHeight) > OutBufSize then
    begin
      Result := False;
      Exit;
    end;
    GetDIBits(dc, bmp, 0, decbmi.bmiHeader.biHeight, OutBuf, decbmi, DIB_RGB_COLORS);
  finally
    DeleteObject(bmp);
    ReleaseDC(wnd, dc);
  end;
  Result := True;
end;

constructor TDibFilter.Create;
begin
  FName := '(Internal DIB to DIB filter)';
end;

function TDibFilter.Test(Buf: PChar; BufSize: Longint): Boolean;
begin
  if BufSize < (SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader)) then
    Result := False
  else if (Buf[0] = 'B') and (Buf[1] = 'M') then
  begin
    Result := IsOS2Dib(Buf)
      or (PBitmapInfoHeader(@Buf[SizeOf(TBitmapFileHeader)]).biSize = SizeOf(TBitmapInfoHeader))
      or (PBitmapInfoHeader(@Buf[SizeOf(TBitmapFileHeader)]).biSize = SizeOf(TBitmapInfoHeader)+16);
        // Photoshop Elementso͂BITFIELDS bmp͂ȂbiSize56
  end else
    Result := False;
end;

function TDibFilter.GetInfo(Buf: PChar; BufSize: Longint; var ImgInfo: TImgInfo): Boolean;
  // 1̃rbg𐔂
  function CountBits(Value: Integer): Integer;
  begin
    Result := 0;
    repeat
      Inc(Result, Value and 1);
      Value := Value shr 1
    until Value = 0;
  end;
var
  bmi: PBitmapInfo;
  bci: PBitmapCoreInfo;
  bf: array[0..2] of DWORD;
  s: string;
begin
  with ImgInfo do
  begin
    if IsOS2Dib(Buf) then
    begin
      bci := @Buf[SizeOf(TBitmapFileHeader)];
      Str(bci.bmciHeader.bcBitCount, s);
      FormatName := 'OS/2 Bitmap, ' + s + 'bit';
    end else
    begin
      bmi := @Buf[SizeOf(TBitmapFileHeader)];
      case bmi.bmiHeader.biCompression of
        BI_RGB:
          begin
            Str(bmi.bmiHeader.biBitCount, s);
            s := 'RGB' + s;
            if bmi.bmiHeader.biBitCount = 16 then
              s := s + '(555)';
          end;
        BI_RLE8: s := 'RLE8';
        BI_RLE4: s := 'RLE4';
        BI_BITFIELDS:
          begin
            Move(bmi.bmiColors[0], bf, SizeOf(bf));
            Str(CountBits(bf[0])*100+CountBits(bf[1])*10+CountBits(bf[2]), s);
            s := 'BITFIELDS(' + s + ')';
          end;
      else
        s := 'unknown';
      end;
      if bmi.bmiHeader.biHeight < 0 then
        s := s + ', topdown';
      FormatName := 'Windows Bitmap, ' + s;
    end;
    GetMem(bmi, MaxBmiSize);
    GetBitmapInfo(Buf, bmi);
    Width := bmi.bmiHeader.biWidth;
    Height := bmi.bmiHeader.biHeight;
    if Height < 0 then
      Height := -Height;
    Bpp := bmi.bmiHeader.biBitCount;  // TODO: RLE24ɂȂႤ
    FreeMem(bmi);
    MimeType := 'image/bmp';
    Comment := '';
  end;
  Result := True;
end;

function TDibFilter.Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar;
  OutBufSize: Longint; BitmapInfo: PBitmapInfo): Boolean;
var
  pitch: Integer;
  top, btm, tmp: PChar;
begin
  GetBitmapInfo(InBuf, BitmapInfo);
  if IsRLE(InBuf) then
    Result := DecodeRLE(InBuf, InBufSize, OutBuf, OutBufSize)
  else begin
    Inc(InBuf, PBitmapFileHeader(InBuf).bfOffBits);
    if InBufSize < OutBufSize then
      OutBufSize := InBufSize;
    Move(InBuf^, OutBuf^, OutBufSize);
    Result := True;
  end;

  // convert top down bmp
  with BitmapInfo.bmiHeader do
    if biHeight < 0 then
    begin
      biHeight := -biHeight;
      pitch := ((biWidth * biBitCount + 7) shr 3 + 3) and not 3;
      top := OutBuf;
      btm := OutBuf + (OutBufSize div pitch - 1) * pitch;
      GetMem(tmp, pitch);
      while top < btm do
      begin
        Move(top^, tmp^, pitch);
        Move(btm^, top^, pitch);
        Move(tmp^, btm^, pitch);
        Inc(top, pitch);
        Dec(btm, pitch);
      end;
      FreeMem(tmp);
    end;
end;

end.
